//usage: ./ipsevil rom.smc patch.ips
//the output is the return code
//0: ok
//1: broken patch or not an ips patch
//2: scrambled patch; likely damaged, but can also be created by a really weird patcher
//3: empty patch
//4: backwards patch
//5: other bad patch
//nonzero should be rejected
//no error checking is in place; nonexistent files will segfault
//
//method:
//if patch application fails, flag broken patch
//if patch block order is illogical, flag damaged patch
// both flips and lips can apply illogical patches, but not create, and there is no reason
//  for any other tool to create those weird patches either, so damage is more plausible
//if patch does not change any bytes, flag empty patch
//if no bytes were changed to different values, flag backwards patch
//if more than 92% of input file is rewritten (changing a byte to itself counts as rewriting), flag other bad patch
// this will catch patches made from bad input files (including empty files) but
//  valid output files; our hack mods are unlikely to catch those
// legitime hacks are below 40% in my tests
// bad patches go above 96% in my tests
//
//demo world legend continues is 56% rewritten, but otherwise clean; the heuristics could be rewritten to
// detect this one, but that kind of weird items are very rare, and I'd rather err on the side of caution
//a pure vanilla hack is also clean
//
//bps validation is trivial - check that the 12th through 9th bytes from the end are identical to crc32 of clean, unheadered rom
//for smw, 89 d4 9e b1
//doesn't check whether the output is bogus, but those are very rare

#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

//the standard library can be assumed to exist
#include <stddef.h>//size_t, SIZE_MAX

#ifndef SIZE_MAX
#define SIZE_MAX ((size_t)-1)
#endif

struct mem {
	unsigned char * ptr;
	size_t len;
};

enum ipserror {
	ips_ok,//Patch applied or created successfully.
	
	ips_notthis,//The patch is most likely not intended for this ROM.
	ips_thisout,//You most likely applied the patch on the output ROM.
	ips_scrambled,//The patch is technically valid, but seems scrambled or malformed.
	ips_invalid,//The patch is invalid.
	
	ips_16MB,//One or both files is bigger than 16MB. The IPS format doesn't support that. The created
	         //patch contains only the differences to that point.
	ips_identical,//The input buffers are identical.
	
	ips_shut_the_fuck_up_gcc//This one isn't used, it's just to kill a stray comma warning.
};

//Applies the IPS patch in [patch, patchlen] to [in, inlen] and stores it to [out, outlen]. Send the
//  return value in out to ips_free when you're done with it.
enum ipserror ips_apply(struct mem patch, struct mem in, struct mem * out);

//Creates an IPS patch that converts source to target and stores it to patch.
enum ipserror ips_create(struct mem source, struct mem target, struct mem * patch);

//Frees the memory returned in the output parameters of the above. Do not call it twice on the same
//  input, nor on anything you got from anywhere else. ips_free is guaranteed to be equivalent to
//  calling stdlib.h's free() on mem.ptr.
void ips_free(struct mem mem);

//ips_study allows you to detect most patching errors without applying it to a ROM, or even a ROM to
//  apply it to. ips_apply calls ips_study and ips_apply_study, so if you call ips_study yourself,
//  it's recommended to call ips_apply_study to not redo the calculation. ips_free is still
//  required.
struct ipsstudy;
enum ipserror ips_study(struct mem patch, struct ipsstudy * study);
enum ipserror ips_apply_study(struct mem patch, struct ipsstudy * study, struct mem in, struct mem * out);


typedef unsigned char byte;

#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)>(b)?(a):(b))
#define clamp(a,b,c) max(a,min(b,c))

struct ipsstudy {
	enum ipserror error;
	unsigned int outlen_min;
	unsigned int outlen_max;
	unsigned int outlen_min_mem;
};

enum ipserror ips_study(struct mem patch, struct ipsstudy * study)
{
	study->error=ips_invalid;
	if (patch.len<8) return ips_invalid;
	const unsigned char * patchat=patch.ptr;
	const unsigned char * patchend=patchat+patch.len;
#define read8() ((patchat<patchend)?(*patchat++):0)
#define read16() ((patchat+1<patchend)?(patchat+=2,((patchat[-2]<<8)|patchat[-1])):0)
#define read24() ((patchat+2<patchend)?(patchat+=3,((patchat[-3]<<16)|(patchat[-2]<<8)|patchat[-1])):0)
	if (read8()!='P' ||
			read8()!='A' ||
			read8()!='T' ||
			read8()!='C' ||
			read8()!='H')
	{
		return ips_invalid;
	}
	unsigned int offset=read24();
	unsigned int outlen=0;
	unsigned int thisout=0;
	unsigned int lastoffset=0;
	bool w_scrambled=false;
	while (offset!=0x454F46)//454F46=EOF
	{
		unsigned int size=read16();
		if (size==0)
		{
			size=read16();
			if (!size) w_scrambled=true;
			thisout=offset+size;
			read8();
		}
		else
		{
			thisout=offset+size;
			patchat+=size;
		}
		if (offset<lastoffset) w_scrambled=true;
		lastoffset=offset;
		if (thisout>outlen) outlen=thisout;
		if (patchat>=patchend) return ips_invalid;
		offset=read24();
	}
	study->outlen_min_mem=outlen;
	study->outlen_max=0xFFFFFFFF;
	if (patchat+3==patchend)
	{
		unsigned int truncate=read24();
		study->outlen_max=truncate;
		if (outlen>truncate)
		{
			outlen=truncate;
			w_scrambled=true;
		}
	}
	if (patchat!=patchend) return ips_invalid;
	study->outlen_min=outlen;
#undef read8
#undef read16
#undef read24
	study->error=ips_ok;
	if (w_scrambled) study->error=ips_scrambled;
	return study->error;
}

unsigned int wastebytes=0;
unsigned int wastebytesnonrle=0;
unsigned int totalbytes=0;
unsigned int totalbytesoverwritten=0;
unsigned int wasteblocks=0;
unsigned int totalblocks=0;

enum ipserror ips_apply_study(struct mem patch, struct ipsstudy * study, struct mem in, struct mem * out)
{
	out->ptr=NULL;
	out->len=0;
	if (study->error==ips_invalid) return study->error;
#define read8() (*patchat++)//guaranteed to not overflow at this point, we already checked the patch
#define read16() (patchat+=2,((patchat[-2]<<8)|patchat[-1]))
#define read24() (patchat+=3,((patchat[-3]<<16)|(patchat[-2]<<8)|patchat[-1]))
	unsigned int outlen=clamp(study->outlen_min, in.len, study->outlen_max);
	out->ptr=(byte*)malloc(max(outlen, study->outlen_min_mem));
	out->len=outlen;
	
	bool anychanges=false;
	if (outlen!=in.len) anychanges=true;
	
	
	if (out->len>in.len)
	{
		memcpy(out->ptr, in.ptr, in.len);
		memset(out->ptr+in.len, 0, out->len-in.len);
	}
	else memcpy(out->ptr, in.ptr, outlen);
	const unsigned char * patchat=patch.ptr+5;
	unsigned int offset=read24();
	while (offset!=0x454F46)
	{
		unsigned int size=read16();
		if (size==0)
		{
			size=read16();
			if (!size) {}//no clue (fix the change detector if changing this)
			unsigned char b=read8();
			
			unsigned int wastestart;
			signed int wasteend;//otherwise it blows up
			for (wastestart=0;wastestart<size && (out->ptr+offset)[wastestart]==b && offset+wastestart<in.len;wastestart++) {}
			for (wasteend=size-1;wasteend>=0 && (out->ptr+offset)[wasteend]==b && offset+wasteend<in.len;wasteend--) {}
			if (wastestart==size)
			{
				wastebytes+=size;
				wasteblocks++;
			}
			else
			{
				wastebytes+=wastestart;
				wastebytes+=(size-1-wasteend);
			}
			
			totalblocks++;
			totalbytes+=size;
			
			if (offset<in.len) totalbytesoverwritten+=size;
			
			if (size && (out->ptr[offset]!=b || memcmp(out->ptr+offset, out->ptr+offset, size-1))) anychanges=true;
			
			memset(out->ptr+offset, b, size);
		}
		else
		{
			unsigned int wastestart;
			signed int wasteend;
			for (wastestart=0;wastestart<size && (out->ptr+offset)[wastestart]==patchat[wastestart] && offset+wastestart<in.len;wastestart++) {}
			for (wasteend=size-1;wasteend>=0 && (out->ptr+offset)[wasteend]==patchat[wasteend] && offset+wasteend<in.len;wasteend--) {}
			if (wastestart==size)
			{
				wastebytes+=size;
				wastebytesnonrle+=size;
				wasteblocks++;
			}
			else
			{
				wastebytes+=wastestart;
				wastebytesnonrle+=wastestart;
				wastebytes+=(size-1-wasteend);
				wastebytesnonrle+=(size-1-wasteend);
			}
			
			totalblocks++;
			totalbytes+=size;
			
			if (offset<in.len) totalbytesoverwritten+=size;
			
			if (memcmp(out->ptr+offset, patchat, size)) anychanges=true;
			
			memcpy(out->ptr+offset, patchat, size);
			patchat+=size;
		}
		offset=read24();
	}
#undef read8
#undef read16
#undef read24
	
	if (study->outlen_max!=0xFFFFFFFF && in.len<=study->outlen_max) study->error=ips_notthis;//truncate data without this being needed is a poor idea
	if (!anychanges) study->error=ips_thisout;
	return study->error;
}

enum ipserror ips_apply(struct mem patch, struct mem in, struct mem * out)
{
	struct ipsstudy study;
	ips_study(patch, &study);
	return ips_apply_study(patch, &study, in, out);
}
#include <stdio.h>
struct mem readfile(const char * fname)
{
	FILE * f=fopen(fname, "rb");
	fseek(f, 0, SEEK_END);
	struct mem in;
	in.len=ftell(f);
	in.ptr=malloc(in.len);
	fseek(f, 0, SEEK_SET);
	if (fread(in.ptr, 1,in.len, f)) {}//shut up gcc
	fclose(f);
	return in;
}

int verify(const char * romname, const char * patchname)
{
	struct mem in=readfile(romname);
	struct mem patch=readfile(patchname);
	struct mem out;
	enum ipserror e=ips_apply(patch, in, &out);
	if (e==ips_invalid) return 1;
	if (e==ips_scrambled) return 2;
	
//printf("%u\n",e);
//printf("%u %u %u\n",wastebytes,wastebytesnonrle,totalbytes);
//printf("%u %zu\n",totalbytesoverwritten,in.len);
//printf("%u %u\n",wasteblocks,totalblocks);
//printf("%lf\n", (double)totalbytesoverwritten/(double)in.len);
	
	if (!totalbytes) return 3;
	if (e==ips_thisout) return 4;
	if ((double)totalbytesoverwritten/(double)in.len > 0.92) return 5;
	
	return 0;
}

int main(int argc, char * argv[])
{
	//printf("%s\n",argv[2]);
	//const char * ver[]={"good", "dmg", "bad-nul", "bad-rv", "bad-x"};
	//printf("%s\n\n", ver[verify(argv[1], argv[2])]);
	return verify(argv[1], argv[2]);
}










